import turtle as _turtle

from math import sqrt as _sqrt
from math import sin as _sin
from math import cos as _cos
from math import pi as _pi

from typing import Union as _Union

_TNumber = _Union[int, float]
_TLogicalPoint = tuple[_TNumber, _TNumber]
_TRealPoint = tuple[_TNumber, _TNumber]


class ScreenFacade:
    """Screen with a virtual system of coordinates for drawing figures such as lines, polylines, \
    arrows and more. It encapsulates the turtle package functionalities. 
    
    `ScreenFacade` lets the users define a logical (or virtual) screen size by translating the given \
    logical points (a screen position with user virtual coordinates) to a real point (the actual \
    coordinates of the given point).

    Parameters
    ----------
    window_title : str
        Title to display on window. Defaults to "Curvipy".
    background_color : str
        Background color. Can either be a name or a hex color code. Defaults to "#FFFFFF".
    window_width : int or None
        Width of the screen window (in pixels). If None, `window_width` equals to 50% of the \
        display width. Defaults to None.
    window_height : int or None
        Height of the screen window (in pixels). If None, `window_height` equals to 75% of \
        the display height. Defaults to None.
    logical_width : int
        Logical width of the screen. This is the width that the users of `ScreenFacade` class \
        will operate with.
        While `window_width` is the real width of the screen, `logical_width` is a virtual \
        representation of it.
    logical_height : int
        Logical height of the screen. This is the height that the users of `ScreenFacade` class \
        will operate with.
        While `window_height` is the real height of the screen, `logical_height` is a virtual \
        representation of it.
    """

    MIN_DRAWING_SPEED = 1
    MAX_DRAWING_SPEED = 10

    def __init__(
        self,
        window_title: str,
        background_color: str,
        window_width: int,
        window_height: int,
        logical_width: int,
        logical_height: int,
    ):
        # Screen setup
        self.__screen = _turtle.Screen()
        if window_width and window_height:
            self.__screen.setup(window_width, window_height)
        self.__screen.clear()
        self.__screen.title(window_title)
        self.__screen.bgcolor(background_color)
        self.background_color = background_color

        self.logical_width = logical_width
        self.logical_height = logical_height

        # Pen setup
        self.__pen = _turtle.Turtle(visible=False)

        # Performance attributes
        self.__pen_width_cache = None
        self.__pen_color_cache = None

    def get_screen_size(self) -> tuple[int, int]:
        """Returns the real width and height of the screen minus an offset.

        The offset is used to fix `turtle.Screen.window_width()` and \
        `turtle.Screen.window_height()` precision.
        """
        width_offset = 65
        height_offset = 65
        width = self.__screen.window_width()
        height = self.__screen.window_height()
        return width - width_offset, height - height_offset

    def get_real_point(self, logical_point: _TLogicalPoint) -> _TRealPoint:
        """Translates the given logical point to a real point.

        Parameters
        ----------
        logical_point : tuple[int or float, int or float]
            Virtual position to be translated.

        Returns
        -------
        tuple[int or float, int or float]
            Translated real position.
        """
        real_width, real_height = self.get_screen_size()
        return (
            logical_point[0] * (real_width / self.logical_width),
            logical_point[1] * (real_height / self.logical_height),
        )

    def goto_drawing(self, point: _TRealPoint, drawing_speed: int) -> None:
        """Moves pen to the given real point leaving a trace.

        Parameters
        ----------
        point : tuple[int or float, int or float]
            Point (real position) to which the pen will go leaving a trace.
        drawing_speed : int
            Speed at which the pen will move. Integer from 1 to 10.
        """
        self.__pen.down()
        self.__pen.speed(drawing_speed)
        self.__pen.goto(point)

    def goto_without_drawing(self, point: _TRealPoint, drawing_speed: int) -> None:
        """Moves pen to the given real point without leaving a trace.

        Parameters
        ----------
        point : tuple[int or float, int or float]
            Point (real position) to which the pen will go.
        drawing_speed : int
            Speed at which the pen will move. Integer from 1 to 10.
        """
        self.__pen.up()
        self.__pen.speed(drawing_speed)
        self.__pen.goto(point)

    def draw_line(
        self,
        start_point: _TLogicalPoint,
        end_point: _TLogicalPoint,
        line_width: int,
        line_color: str,
        drawing_speed: int,
    ) -> None:
        """Draws a line from start position to end position.

        Parameters
        ----------
        start_point : tuple[int or float, int or float]
            Logical position at which the line starts. `start_point` is \
            translated to a real position.
        end_point : tuple[int or float, int or float]
            Logical position at which the line ends. `end_point` is \
            translated to a real position.
        line_width : int
            Line width.
        line_color : str
            Line color.
        drawing_speed : int
            Drawing speed. Integer from 1 to 10.
        """
        # Pen setup
        if line_width != self.__pen_width_cache:
            self.__pen.width(line_width)
            self.__pen_width_cache = line_width

        if line_color != self.__pen_color_cache:
            self.__pen.color(line_color)
            self.__pen_color_cache = line_color

        # Go to first line point without drawing
        rstart_point = self.get_real_point(start_point)
        self.goto_without_drawing(rstart_point, __class__.MAX_DRAWING_SPEED)
        # Draw line
        rend_point = self.get_real_point(end_point)
        self.goto_drawing(rend_point, drawing_speed)

    def draw_polyline(
        self,
        points: list[_TLogicalPoint],
        polyline_width: int,
        polyline_color: str,
        drawing_speed: int,
    ) -> None:
        """Draws a polyline by joining the given points.

        Parameters
        ----------
        points : list[tuple[int or float, int or float]]
            List of the logical position of the polyline points.
        polyline_width : int
            Polyline width.
        polyline_color : str
            Polyline color.
        drawing_speed : int
            Drawing speed. Integer from 1 to 10.
        """
        # Pen setup
        if polyline_width != self.__pen_width_cache:
            self.__pen.width(polyline_width)
            self.__pen_width_cache = polyline_width

        if polyline_color != self.__pen_color_cache:
            self.__pen.color(polyline_color)
            self.__pen_color_cache = polyline_color

        # Variables
        rpoints = [self.get_real_point(point) for point in points]

        # Go to first polyline point without drawing
        self.__pen.speed(__class__.MAX_DRAWING_SPEED)
        self.__pen.up()
        self.__pen.goto(rpoints[0])
        # Draw polyline
        self.__pen.speed(drawing_speed)
        self.__pen.down()
        for rpoint in rpoints[1:]:
            self.__pen.goto(rpoint)

    def draw_arrow(
        self,
        point: _TLogicalPoint,
        arrow_size: int,
        arrow_angle: _TNumber,
        arrow_width: int,
        arrow_color: str,
        drawing_speed: int,
    ) -> None:
        """Draws an arrow `>` at the given point.

        Parameters
        ----------
        point : tuple[int or float, int or float]
            Logical position where the arrow will be drawn.
        arrow_size : int
            Size of the arrow.
        arrow_angle : str
            Angle in radians. Indicates arrow direction.
        arrow_width : int
            Width of the arrow.
        arrow_color : str
            Color of the arrow.
        drawing_speed : int
            Drawing speed. Integer from 1 to 10.
        """
        # Pen setup
        self.__pen.width(arrow_width)
        self.__pen.color(arrow_color)

        # Variables setup
        rpoint = self.get_real_point(point)

        # Draw vector head
        left_arrow = (
            arrow_size * _cos(arrow_angle + _pi * 5 / 4),
            arrow_size * _sin(arrow_angle + _pi * 5 / 4),
        )
        left_arrow_endpoint = (
            rpoint[0] + left_arrow[0],
            rpoint[1] + left_arrow[1],
        )

        self.goto_without_drawing(rpoint, __class__.MAX_DRAWING_SPEED)
        self.goto_drawing(left_arrow_endpoint, drawing_speed)

        right_arrow = (
            arrow_size * _cos(arrow_angle - _pi * 5 / 4),
            arrow_size * _sin(arrow_angle - _pi * 5 / 4),
        )
        right_arrow_endpoint = (
            rpoint[0] + right_arrow[0],
            rpoint[1] + right_arrow[1],
        )

        # Draw arrow left size
        self.goto_without_drawing(rpoint, __class__.MAX_DRAWING_SPEED)
        self.goto_drawing(right_arrow_endpoint, drawing_speed)

    def draw_text(
        self,
        text: str,
        point: _TLogicalPoint,
        text_font: tuple[str, str, str],
        text_color: str,
        align: str,
    ) -> None:
        """Display the given text at the specified point.

        Parameters
        ----------
        text : str
            Text to be displayed.
        point : tuple[int or float, int or float]
            Logical position where the text will be drawn.
        text_font : tuple[str, str, str]
            A triple (fontname, fontsize, fonttype).
        text_color : str
            Text color.
        align : str
            Text alignment.
        """
        # Pen setup
        if text_color != self.__pen_color_cache:
            self.__pen.color(text_color)
            self.__pen_color_cache = text_color

        # Draw text
        rpoint = self.get_real_point(point)
        self.__pen.speed(__class__.MAX_DRAWING_SPEED)
        self.__pen.up()
        self.__pen.goto(rpoint)
        self.__pen.write(text, move=False, align=align, font=text_font)

    def clean(self) -> None:
        """Removes all drawings from screen."""
        self.__screen.clear()
        self.__screen.bgcolor(self.background_color)

    def exit_on_click(self):
        """Exits screen when clicked."""
        self.__screen.exitonclick()



from math import ceil as _ceil
from math import pi as _pi

from typing import Union as _Union

from ._vector import Vector as _Vector
from ._curve import Curve as _Curve
from ._screen import ScreenFacade as _ScreenFacade

_TNumber = _Union[int, float]


class ScreenConfiguration:
    """Defines the configuration for the plotter screen.

    Parameters
    ----------
    window_title : str
        Title to display on window. Defaults to "Curvipy".
    background_color : str
        Background color. Can either be a name or a hex color code. Defaults to "#FFFFFF".
    window_width : int or None
        Width of the screen window (in pixels). If None, `window_width` equals to 50% of the \
        display width.
    window_height : int or None
        Height of the screen window (in pixels). If None, `window_height` equals to 75% of \
        the display height.
    """

    def __init__(
        self,
        window_title: str = "Curvipy",
        background_color: str = "#FFFFFF",
        window_width: int = None,
        window_height: int = None,
    ):
        self.window_title = window_title
        self.background_color = background_color
        self.window_width = window_width
        self.window_height = window_height


class PlottingConfiguration:
    """Defines the configuration for plotting curves and vectors.

    Parameters
    ----------
    plotting_speed : int
        Curve plotting speed. Integer from 1 to 10. Defaults to 10.
    curve_color : str
        Curve color. Can either be a name or a hex color code. Defaults to "#457B9D".
    curve_width : int
        Curve width. Defaults to 4.
    vector_color : str
        Vector color. Can either be a name or a hex color code. Defaults to "#E63946".
    vector_width : int
        Vector width. Defaults to 3.
    vector_head_size : int
        Size of vectors' head. Defaults to 10.
    """

    def __init__(
        self,
        plotting_speed: int = 10,
        curve_color: str = "#457B9D",
        curve_width: int = 4,
        vector_color: str = "#E63946",
        vector_width: int = 3,
        vector_head_size: int = 10,
    ):
        # General attributes
        self.plotting_speed = plotting_speed

        # Curves attributes
        self.curve_color = curve_color
        self.curve_width = curve_width

        # Vectors
        self.vector_color = vector_color
        self.vector_width = vector_width
        self.vector_head_size = vector_head_size


class AxesConfiguration:
    """Defines the configuration for the plotter axes.
        
    Parameters
    ----------
    show_axes : bool
        If True, x-axis and y-axis are shown. Defaults to True.
    show_axes_direction : bool
        If True, `Plotter` replaces the last tick of both axes with an arrow that \
        indicates the axis direction. Defaults to False.
    axes_color : str
        Axis color. Can either be a name  or a hex color code. Defaults to "#70CBCE".
    axes_width : int
        Axis width. Defaults to 2.
    ticks_font : tuple[str, str, str]
        A triple (fontname, fontsize, fonttype). Defaults to ("Verdana", 8, "normal"). 
    ticks_color : str
        Ticks text color. Can either be a name or a hex color code. Defaults to "#000000".
    x_ticks : int
        Number of ticks of half the x-axis, i.e. total x-axis ticks equals `2 * x_ticks`. \
        Defaults to 10.
    x_ticks_distance : int
        Distance between each x-axis tick. E.g. if `x_ticks_distance` equals 1, x-axis \
        ticks will be `{..., -2, -1, 0, 1, 2, ...}`. Defaults to 1.
    x_ticks_decimals : int
        Number of decimals of the x-axis ticks number. Defaults to 2.
    x_ticks_align : str
        Can either be "top" or "down". Defines if the x-axis ticks number will be placed \
        upside or downside the x-axis. Defaults to "down".
    x_ticks : int
        Number of ticks of half the y-axis, i.e. total y-axis ticks equals `2 * y_ticks`. \
        Defaults to 10.
    y_ticks_distance : int
        Distance between each y-axis tick. E.g. if `y_ticks_distance` equals 0.5, y-axis \
        ticks will be `{..., -1, -0.5, 0, 0.5, 1, ...}`. Defaults to 1.
    y_tick_decimals : int
        Number of decimals of the y-axis ticks number. Defaults to 2.
    y_ticks_align : str
        Can either be "left" or "right". Defines if the y-axis ticks number will be placed \
        to the left or to the right of the y-axis. Defaults to "left".
    """

    def __init__(
        self,
        show_axes: bool = True,
        show_axes_direction: bool = False,
        axes_color: str = "#70CBCE",
        axes_width: int = 2,
        ticks_font: tuple[str, str, str] = ("Verdana", 8, "normal"),
        ticks_color: str = "#000000",
        x_ticks: int = 10,
        x_ticks_distance: int = 1,
        x_ticks_decimals: int = 2,
        x_ticks_align: str = "down",
        y_ticks: int = 10,
        y_ticks_distance: int = 1,
        y_ticks_decimals: int = 2,
        y_ticks_align: str = "left",
    ):
        # General attributes
        self.show_axes = show_axes
        self.show_axes_direction = show_axes_direction
        self.axes_color = axes_color
        self.axes_width = axes_width
        self.ticks_font = ticks_font
        self.ticks_color = ticks_color
        # X-axis ticks attributes
        self.x_ticks = x_ticks
        self.x_ticks_distance = x_ticks_distance
        self.x_ticks_decimals = x_ticks_decimals
        self.x_ticks_align = x_ticks_align
        # Y-axis ticks attributes
        self.y_ticks = y_ticks
        self.y_ticks_distance = y_ticks_distance
        self.y_ticks_decimals = y_ticks_decimals
        self.y_ticks_align = y_ticks_align


class Plotter:
    """Graph plotter for drawing curves and vectors.

    Parameters
    ----------
    screen_config : ScreenConfiguration
        Screen configuration. By default, `screen_config` takes the defaults attributes \
        values of `ScreenConfiguration`.
    plotting_config : PlottingConfiguration
        Plotting configuration. By default, `plotting_config` takes the defaults attributes \
        values of `PlottingConfiguration`.
    axes_config : AxesConfiguration
        Axes configuration.  By default, `axes_config` takes the defaults attributes values \
        of `AxesConfiguration`.
    """

    def __init__(
        self,
        screen_config: ScreenConfiguration = None,
        plotting_config: PlottingConfiguration = None,
        axes_config: AxesConfiguration = None,
    ) -> None:
        # Plotter coordinates system:
        #
        # `curvipy.Plotter` works with a virtual set of coordinates defined by the a logical width
        # and height. The logical screen size is independent of the real window size, and virtual
        # coordinates are translated to real window coordinates.
        # Despite setting a logical width and height do not affect the window size, it defines the
        # xy-plane coordinate system. This is useful for precisely calibrating the distance between
        # the axes ticks.
        #
        # Note that the x-axis length equals the logical width and the y-axis length equals the
        # logical height.
        #
        # If we want the x-axis to have `n` total ticks with a distance `Dx` between their self, that is
        #
        #                           axis_ticks = {Dx. i | i in N ^ -n/2 <= i <= n/2}
        #                                      = {..., -2.Dx, -Dx, 0, Dx, 2.Dx, ....}
        #
        # and the y-axis to have `m` total ticks with a distance `Dy` between their self, that is
        #
        #                           axis_ticks = {Dy. i | i in N ^ -m/2 <= i <= m/2}
        #                                      = {..., -2.Dy, -Dy, 0, Dy, 2.Dy, ....}
        #
        # then
        #
        #                                   logical_width = n . Dx
        #                                   logical_height = m . Dy
        #

        # Public attributes
        if screen_config is not None:
            self.screen_config = screen_config
        else:
            self.screen_config = ScreenConfiguration()

        if plotting_config is not None:
            self.plotting_config = plotting_config
        else:
            self.plotting_config = PlottingConfiguration()

        if axes_config is not None:
            self.axes_config = axes_config
        else:
            self.axes_config = AxesConfiguration()

        # Private attributes
        total_x_ticks = 2 * self.axes_config.x_ticks
        total_y_ticks = 2 * self.axes_config.y_ticks

        self.__logical_width = total_x_ticks * self.axes_config.x_ticks_distance
        self.__logical_height = total_y_ticks * self.axes_config.y_ticks_distance

        self.__screen = _ScreenFacade(
            self.screen_config.window_title,
            self.screen_config.background_color,
            self.screen_config.window_width,
            self.screen_config.window_height,
            self.__logical_width,
            self.__logical_height,
        )

        if self.axes_config.show_axes:
            self._draw_axis()

    def _draw_axis(self) -> None:
        w, h = self.__logical_width, self.__logical_height

        # Draw y-axis line
        self.__screen.draw_line(
            (0, -h / 2),
            (0, h / 2),
            self.axes_config.axes_width,
            self.axes_config.axes_color,
            self.__screen.MAX_DRAWING_SPEED,
        )

        # Draw y-axis ticks
        dy = 0 if not self.axes_config.y_ticks else h / (self.axes_config.y_ticks * 2)
        for i in range(self.axes_config.y_ticks * 2):
            y = dy * (i - self.axes_config.y_ticks)
            if y != 0:
                self.draw_y_tick(y)

        # Draw y-axis last tick or arrow (see `AxesConfiguration.show_axes_direction` attribute description)
        if self.axes_config.show_axes_direction:
            axis_arrow_pos = (0, h / 2)
            axis_arrow_ang = 0.5 * _pi
            self.__screen.draw_arrow(
                point=axis_arrow_pos,
                arrow_angle=axis_arrow_ang,
                arrow_size=10,
                arrow_width=self.axes_config.axes_width,
                arrow_color=self.axes_config.axes_color,
                drawing_speed=self.__screen.MAX_DRAWING_SPEED,
            )

        else:
            last_tick = dy * self.axes_config.y_ticks
            self.draw_y_tick(last_tick)

        # Draw x-axis line
        self.__screen.draw_line(
            (-w / 2, 0),
            (w / 2, 0),
            self.axes_config.axes_width,
            self.axes_config.axes_color,
            self.__screen.MAX_DRAWING_SPEED,
        )

        # Draw x-axis ticks
        dx = 0 if not self.axes_config.x_ticks else w / (self.axes_config.x_ticks * 2)
        for i in range(self.axes_config.x_ticks * 2):
            x = dx * (i - self.axes_config.x_ticks)
            if x != 0:
                self.draw_x_tick(x)

        # Draw x-axis last tick or arrow (see `AxesConfiguration.show_axes_direction` attribute description)
        if self.axes_config.show_axes_direction:
            axis_arrow_pos = (w / 2, 0)
            axis_arrow_ang = 0
            self.__screen.draw_arrow(
                point=axis_arrow_pos,
                arrow_angle=axis_arrow_ang,
                arrow_size=10,
                arrow_width=self.axes_config.axes_width,
                arrow_color=self.axes_config.axes_color,
                drawing_speed=self.__screen.MAX_DRAWING_SPEED,
            )
        else:
            last_tick = dx * self.axes_config.x_ticks
            self.draw_x_tick(last_tick)

    def draw_x_tick(self, number: _TNumber, align: str = None) -> None:
        """Draws a tick on the x-axis.

        Parameters
        ----------
        number : int or float
            Tick number.
        align : str
            Can either be "top" or "down". Defines if the x-axis tick number will be placed \
            upside or downside the y-axis. If `None`, align takes \
            `Plotter.x_ticks_align` attribute value. Defaults to `None`.
        """
        # NOTE: values for `y_offset`, `text_offset` and `tick_length` have been determined
        # experimentally. This method work properly for small values of text font size
        # and well spaced ticks. If the given font is too big or ticks are drawn relative
        # closely, this method might overlap the ticks with their self or with the axes.

        # Variables setup
        w, h = self.__logical_width, self.__logical_height
        font_size = self.axes_config.ticks_font[1]
        number_str = str(round(number, self.axes_config.x_ticks_decimals))

        # Offsets
        align = self.axes_config.x_ticks_align if not align else align
        if align == "down":
            y_offset = -0.04 * h  # -4% of the y-axis length
        if align == "up":
            y_offset = 0.02 * h  # 2% of the y-axis length

        text_offset = -font_size * len(number_str) * w * 0.5e-3

        # Draw x tick
        tick_length = 0.01 * h  # 1% of the y-axis length
        self.__screen.draw_line(
            (number, -tick_length),
            (number, tick_length),
            self.axes_config.axes_width,
            self.axes_config.axes_color,
            self.__screen.MAX_DRAWING_SPEED,
        )
        self.__screen.draw_text(
            text=number_str,
            point=(number + text_offset, 0 + y_offset),
            text_font=self.axes_config.ticks_font,
            text_color=self.axes_config.ticks_color,
            align="left",
        )

    def draw_y_tick(self, number: _TNumber, align: str = None) -> None:
        """Draws a tick on the y-axis.

        Parameters
        ----------
        number : int or float
            Tick number.
        align : str
            Can either be "left" or "right". Defines if the y-axis ticks number will be placed \
            to the left or to the right of the x-axis. If `None`, align takes \
            `Plotter.y_ticks_align` attribute value. Defaults to `None`.
        """
        # NOTE: values for `y_offset`, `x_offset`, `tick_length` and `tick_length` have been
        # determined experimentally. This method work properly for small values of text
        # font size and well spaced ticks. If the given font is too big or ticks are drawn
        # relative closely, this method might overlap the ticks with their self or with the
        # axes.

        # Variables setup
        w, h = self.__logical_width, self.__logical_height
        font_size = self.axes_config.ticks_font[1]
        number_str = str(round(number, self.axes_config.y_ticks_decimals))

        # Offsets
        align = self.axes_config.y_ticks_align if not align else align
        if align == "left":
            text_offset = font_size * len(number_str) * w * 1e-3
            x_offset = -0.02 * w - text_offset  # 2% of x-axis length plus text offset
        if align == "right":
            x_offset = 0.02 * w  # 2% of x-axis length

        y_offset = -font_size * h * 1e-3

        # Draw y tick
        tick_length = 0.01 * w  # 1% of the x-axis length
        self.__screen.draw_line(
            (-tick_length, number),
            (tick_length, number),
            self.axes_config.axes_width,
            self.axes_config.axes_color,
            self.__screen.MAX_DRAWING_SPEED,
        )
        self.__screen.draw_text(
            text=number_str,
            point=(0 + x_offset, number + y_offset),
            text_font=self.axes_config.ticks_font,
            text_color=self.axes_config.ticks_color,
            align="left",
        )

    def plot_vector(self, vector: _Vector) -> None:
        """Plots the given two-dimensional vector.

        Parameters
        ----------
        vector : Vector
            Vector to be plotted.
        """
        # NOTE: for plotting the vector arrow, we cannot use the vector angle because
        # the vector components do not reflect the real coordinates where it will be
        # drawn. Because of that, we need the angle of a scaled version of itself, that
        # reflect the real coordinates on screen of the vector.
        #
        # TODO:
        #   - Display vector components

        # Check if vector is the cero vector (v = [0, 0])
        if not vector.norm:
            return

        # Draw vector
        self.__screen.draw_line(
            vector.tail,
            vector.head,
            self.plotting_config.vector_width,
            self.plotting_config.vector_color,
            self.plotting_config.plotting_speed,
        )

        # Draw vector head
        scaled_vector = _Vector(self.__screen.get_real_point(vector.components))
        self.__screen.draw_arrow(
            point=vector.head,
            arrow_angle=scaled_vector.angle,
            arrow_size=self.plotting_config.vector_head_size,
            arrow_width=self.plotting_config.vector_width,
            arrow_color=self.plotting_config.vector_color,
            drawing_speed=self.__screen.MAX_DRAWING_SPEED,
        )

    def plot_curve(self, curve: _Curve) -> None:
        """Plots the given two-dimensional curve in the specified interval.

        Parameters
        ----------
        curve : Curve
            Curve to be plotted.
        """
        # Draw curve
        curve_points = curve.points()
        self.__screen.draw_polyline(
            curve_points,
            self.plotting_config.curve_width,
            self.plotting_config.curve_color,
            self.plotting_config.plotting_speed,
        )

    def plot_animated_curve(self, curve: _Curve, samples_per_vector: int) -> None:
        """Plots the given curve by drawing a set of vectors pointing at the curve \
        points and then joining the vector heads.

        Parameters
        ----------
        curve : Curve
            Curve to be plotted.
        samples_per_vector : int
            Number of samples per each vector plotted. The less `samples_per_vector` \
            is, the more vectors are drawn. 
        """
        # Plot vectors:
        for i, vector in enumerate(curve.points()):
            if i % samples_per_vector != 0:
                continue
            self.plot_vector(_Vector(vector))

        # Plot curve:
        self.plot_curve(curve)

    def clean(self) -> None:
        """Removes curves and vectors plotted."""
        self.__screen.clean()
        if self.axes_config.show_axes:
            self._draw_axis()

    def wait(self) -> None:
        """Waits until plotter screen is clicked. When clicked, exits plotter."""
        self.__screen.exit_on_click()



from abc import ABC as _ABC
from abc import abstractmethod as _abstractmethod

from typing import Callable as _Callable
from typing import Union as _Union

from ._interval import Interval as _Interval

_TNumber = _Union[int, float]
_TVector = tuple[_TNumber, _TNumber]


class Curve(_ABC):
    """Base class for all two-dimensional curves."""

    @_abstractmethod
    def points(self) -> list[_TNumber]:
        """Returns a sorted list of curve point. The position of the points \
        indicates the order in which they will be plotted. 

        Returns
        -------
        list[int or float]
            A list of curve points. The position of the points indicates the \
            order in which they will be plotted.
        """
        pass


class Function(Curve):
    """Function that given a real number returns another real number `y = f(x)`.

    Parameters
    ----------
    function : Callable[[int or float], int or float]
        Function that given an integer or float returns another integer or float.
    interval : Interval
            The interval from which the curve will be plotted.
    """

    def __init__(self, function: _Callable[[_TNumber], _TNumber], interval: _Interval):
        self.function = function
        self.interval = interval

    def points(self) -> list[_TNumber]:
        """Returns the function point for each value in the given interval.

        Returns
        -------
        list[int or float]
            A list of function points.
        """
        return [(x, self.function(x)) for x in self.interval]


class ParametricFunction(Curve):
    """Function that given a real number returns a 2-dimensional vector `f(t) = <x(t), y(t)>`.

    Parameters
    ----------
    function : Callable[[int or float], tuple[int or float, int or float]]
                Function that given an integer or float returns a tuple containing two \
                integers or floats.
    interval : Interval
            The interval from which the parametric function will be plotted.
    """

    def __init__(
        self, parametric_function: _Callable[[_TNumber], _TVector], interval: _Interval
    ):
        self.parametric_function = parametric_function
        self.interval = interval

    def points(self) -> list[_TNumber]:
        """Returns the parametric function point for each value in the given interval.

        Returns
        -------
        list[int or float]
            A list of parametric function points.
        """
        return [self.parametric_function(t) for t in self.interval]


class TransformedCurve(Curve):
    """Applies a linear transformation (defined as a 2x2 matrix) to the given curve.

    Parameters
    ----------
    matrix : tuple[\
            tuple[int or float, int or float],\
            tuple[int or float, int or float]\
            ]
        Linear transformation represented as a 2x2 matrix.
        
    curve : Curve
        Curve to be transformed.
    """

    def __init__(
        self,
        matrix: tuple[
            tuple[_TNumber, _TNumber],
            tuple[_TNumber, _TNumber],
        ],
        curve: Curve,
    ):
        self.__curve = curve
        self.__matrix = matrix

    def points(self) -> list[_TNumber]:
        """Applies a linear transformation to each point of the curve.

        Returns
        -------
        list[int or float]
            A list of transformed curve points.
        """
        points = []
        for point in self.__curve.points():
            points.append(
                [
                    point[0] * self.__matrix[0][0] + point[1] * self.__matrix[0][1],
                    point[0] * self.__matrix[1][0] + point[1] * self.__matrix[1][1],
                ]
            )
        return points



import math as _math

from typing import Union as _Union

_TNumber = _Union[int, float]
_TPoint = tuple[_TNumber, _TNumber]
_TVector = tuple[_TNumber, _TNumber]


class Vector:
    """Two-dimensional vector

        .. math::
            \\vec{v} = (v_1, v_2)

    Attributes
    ----------
    head : tuple[int or float, int or float]
        Two-dimensional point at which the vector is pointing.
    tail : tuple[int or float, int or float]
        Two-dimensional point at which the vector starts.
    """

    def __init__(
        self,
        head: _TPoint,
        tail: _TPoint = (0, 0),
    ):
        self.head = head
        self.tail = tail

    @property
    def components(self) -> _TVector:
        """Vector head point when tail is moved to the origin."""
        return (self.head[0] - self.tail[0], self.head[1] - self.tail[1])

    @property
    def norm(self) -> float:
        """Norm of the vector."""
        return _math.sqrt(self.components[0] ** 2 + self.components[1] ** 2)

    @property
    def angle(self) -> float:
        """Angle between the vector and the x-axis in radians. This angle \
        is a real number in the interval [-π, π]. If the angle is positive \
        it indicates an anticlockwise rotation from the x-axis, this means \
        that the vector is on the first or the second quadrant. If the angle \
        is negative it indicates a clockwise rotation from the x-axis, this \
        means that the vector is on the third or fourth quadrant.
        """
        angle = _math.acos(self.components[0] / self.norm)
        return angle if self.components[1] >= 0 else -angle

    def place(self, point: _TPoint) -> None:
        """Moves the vector to the given point, that is, the vector tail \
        is placed on the specified coordinates.

        Parameters
        ----------
        point : tuple[int or float, int or float]
            Two-dimensional point to which the vector is moved.

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=(-1, -1), head=(2, 2))
            v.place((-2, 2))
            v.tail
            >>> (-2, 2)
            v.head
            >>> (1, 5)
        """
        self.head = (
            point[0] + self.components[0],
            point[1] + self.components[1],
        )
        self.tail = point

    def __getitem__(self, i: int) -> _TNumber:
        """Returns the component at index `i`.

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=[-1, -1], head=[2, 3])
            v.components
            >>> (3, 4)
            v[0]
            >>> 3
            v[1]
            >>> 4
        """
        return self.components[i]

    def __len__(self) -> int:
        return len(self.components)

    def __mul__(self, scalar: _TNumber) -> "Vector":
        """Defines vector scaling. The scaled vector preserves the \
        vector tail.

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=[-1, -1], head=[2, 2])
            w = v * 2
            w.tail
            >>> (-1, -1)
            w.head
            >>> (5, 5)
            w.components
            >>> (6, 6)
        """
        scaled_vector_components = (
            self.components[0] * scalar,
            self.components[1] * scalar,
        )
        scaled_vector_head = (
            self.tail[0] + scaled_vector_components[0],
            self.tail[1] + scaled_vector_components[1],
        )
        return Vector(head=scaled_vector_head, tail=self.tail)

    def __add__(self, vector: "Vector") -> "Vector":
        """Defines vector addition. Result vector is placed at the origin, \
        that is, the tail of the result vector is (0, 0).

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=[-1, -1], head=[2, 2])
            w = curvipy.Vector(tail=[2, 0], head=[1, -2])
            n = v + w
            n.tail
            >>> (0, 0)
            n.head
            >>> (2, 1)"""
        head = (
            self.components[0] + vector.components[0],
            self.components[1] + vector.components[1],
        )
        return Vector(head)

    def __sub__(self, vector: "Vector") -> "Vector":
        """Defines vector subtraction. Result vector is placed at the origin, \
        that is, the tail of the result vector is (0, 0).

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=[-1, -1], head=[2, 2])
            w = curvipy.Vector(tail=[2, 0], head=[1, -2])
            n = v - w
            n.tail
            >>> (0, 0)
            n.head
            >>> (4, 5)
        """
        return self + vector * -1

    def __eq__(self, vector: "Vector") -> bool:
        """Defines equality between two vectors. It compares the components \
        of two vectors.

        Example
        -------
        .. code-block:: python

            v = curvipy.Vector(tail=[1, 1], head=[3, 3])
            w = curvipy.Vector(tail=[0, 0], head=[2, 2])
            v == w
            >>> True
        """
        return self.components == vector.components
